﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Sandbox.Data;
using Sandbox.IO;
using Sandbox.Forms;
using System.ComponentModel;
using System.Xml.Serialization;

namespace Sandbox.Usermap
{
    public class H3Usermap
    {
        private EndianIO io;
        public Header header;
        public UnnamedBlock1 unamedBlock1;
        public PlacementBlock[] placementBlocks;
        public TagIndexEntry[] tagIndex;

        public int baseObjectCount;

        public H3Usermap() { }

        public H3Usermap(string filepath)
        {
            io = new EndianIO(filepath, EndianType.BigEndian);
            io.Open();
        }
        /// <summary>
        /// Closes currently open usermap.
        /// </summary>
        public void Close()
        {
            io.Close();
        }
        /// <summary>
        /// Reads usermap into memory
        /// </summary>
        /// <returns>Error code</returns>
        public int Read()
        {
            try
            {
                //Start at beggining of file.
                io.In.BaseStream.Position = 0;
                //Read the file type string.
                if (new string(io.In.ReadChars(4)) == "CON ")
                    //Returns error code 1:
                    // Usermap must be extrcted from container.
                    return 1; 

                //Move past blf_ and chdr headers.
                io.In.BaseStream.Position = 312;

                //Read file header string.
                if (new string(io.In.ReadChars(4)) != "mapv")
                    //Returns error code 2:
                    // This is not a usermap file.
                    return 2;

                //Move back to the header.
                io.In.BaseStream.Position = 312;
                header = new Header(); //Create new instance.
                header.Read(io.In); //Read header.  

                unamedBlock1 = new UnnamedBlock1();
                unamedBlock1.Read(io.In);

                //Move to start of placement blocks.
                io.In.BaseStream.Position = 632;
                //Create new array to hold blocks.
                int objCount = 0;
                placementBlocks = new PlacementBlock[640];
                for (int i = 0; i < 640; i++) //Loop through blocks.
                {
                    placementBlocks[i] = new PlacementBlock(); //Creat new instance.
                    placementBlocks[i].Read(io.In); //Read block.
                    /*if (placementBlocks[i].TagIndex >= 25 && placementBlocks[i].TagIndex <= 28)
                        placementBlocks[i].Z += 1;*/
                    if (placementBlocks[i].BlockType != PlacementBlock.blockType.NULL)
                        objCount++;
                }
                baseObjectCount = unamedBlock1.Spawned_Object_Count - objCount;
                //Move to start iof tag index.
                io.In.BaseStream.Position = 54420;
                //Creat new array to hold index entrys.
                tagIndex = new TagIndexEntry[256];
                for (int i = 0; i < 256; i++) //Loop through entrys.
                {
                    tagIndex[i] = new TagIndexEntry(); //Create new instance.
                    tagIndex[i].Read(io.In); //Read entry.
                } 
            }
            catch
            {
                //Returns error code 3:
                // An error was encountered reading usermap.
                return 3;
            }

            //Returns error code 0:
            // Success.
            return 0;
        }
        /// <summary>
        /// Write usermap from memory to file
        /// </summary>
        /// <returns>Error code</returns>
        public int Write()
        {
            try
            {
                io.Out.BaseStream.Position = 312;
                header.Write(io.Out);

                unamedBlock1.Write(io.Out);

                io.Out.BaseStream.Position = 632;
                for (int i = 0; i < 640; i++)
                {
                    placementBlocks[i].Write(io.Out);
                }                

                io.Out.BaseStream.Position = 54420;
                for (int i = 0; i < 256; i++)
                {
                    tagIndex[i].Write(io.Out);
                }
            }
            catch
            {
                return 1;
            }

            return 0;
        }

        public int FindPlacementBlock(int Ident, int index)
        {
            for (int i = index; i < 640; i++)
            {
                int tagIndex = GlobalVariables.Usermap.placementBlocks[i].TagIndex;
                if ((tagIndex != -1) && (GlobalVariables.Usermap.tagIndex[tagIndex].ident == Ident))
                {
                    return i;
                }
            }
            return -1;
        }
        /// <summary>
        /// Reads, writes, and allows editing of the usermap header..
        /// </summary>
        public class Header : Chunk
        {
            //Structure of usermap header
            #region Structure
            [Size(4)]
            public char[] MapV;
            public int Size;
            public short Unknown_8;
            public short Unknown_10;
            [Size(12)]
            public byte[] Unknown_12;
            [String(StringType.Unicode, 16)]
            public string varientName;
            [String(128)]
            public string varientDescription;
            [String(16)]
            public string varientAuthor;
            public int Unknown_200;
            public int Unknown_204;
            public int Unknown_208;
            public int Unknown_212;
            public int Unknown_216;
            public int Entry_Size;
            public int CON_File_Name_1;
            public int CON_FIle_Name_2;
            public int Unknown_232;
            public int NULL;
            public int varientMapId;
            public int Unknown_244;
            public int NULL_2;
            public short NULL_3;
            [Size(10)]
            public byte[] Unknown_254;
            #endregion
            //Property's for values to allow editing of later.
            #region Properties
            [Category("Map Info"), System.ComponentModel.Description("Author of this usermap")]
            public string Author
            {
                get { return varientAuthor; }
                set { varientAuthor = OnAuthorChange(value); }
            }
            [Category("Map Info"), System.ComponentModel.Description("Name of this usermap.")]
            public string Name
            {
                get { return varientName; }
                set { varientName = OnNameChange(value); }
            }
            [Category("Map Info"), System.ComponentModel.Description("Desscription of this usermap.")]
            public string Description
            {
                get { return varientDescription; }
                set { varientDescription = OnDescriptionChange(value); }
            }
            [Category("Map Info"), System.ComponentModel.Description("What map is this usermap for.")]
            public Misc.MapID MapId
            {
                get { return (Misc.MapID)varientMapId; }
                set { varientMapId = (int)value; }
            }
            #endregion
            //Methods for perfoming operations on placement blocks.
            #region Methods
            protected string OnAuthorChange(string author)
            {
                if (author.Length <= 16)
                    return author;
                else
                    System.Windows.Forms.MessageBox.Show("Author name must not be greater than 16 characters.",
                        "Too long.", System.Windows.Forms.MessageBoxButtons.OK, System.Windows.Forms.MessageBoxIcon.Error);
                return varientAuthor;
            }
            protected string OnNameChange(string name)
            {
                if (name.Length <= 16)
                    return name;
                else
                    System.Windows.Forms.MessageBox.Show("Map name must not be greater than 16 characters.",
                        "Too long.", System.Windows.Forms.MessageBoxButtons.OK, System.Windows.Forms.MessageBoxIcon.Error);
                return name;
            }
            protected string OnDescriptionChange(string description)
            {
                if (description.Length <= 128)
                    return description;
                else
                    System.Windows.Forms.MessageBox.Show("Description name must not be greater than 128 characters.",
                        "Too long.", System.Windows.Forms.MessageBoxButtons.OK, System.Windows.Forms.MessageBoxIcon.Error);
                return varientDescription;
            }
            #endregion
        }
        /// <summary>
        /// Reads, writes, and allows editing of unknown block 1
        /// </summary>
        public class UnnamedBlock1 : Chunk //56 Bytes
        {
            //Structure of unamed block 1
            #region Structure
            public short Unknown;
            public short Unknown_2;
            public byte Unknown_4;
            public byte Spawned_Object_Count;
            public short Unknown_6;
            public int Map_ID;
            public float World_Bounds_X_Min;
            public float World_Bounds_X_Max;
            public float World_Bounds_Y_Min;
            public float World_Bounds_Y_Max;
            public float World_Bounds_Z_Min;
            public float World_Bounds_Z_Max;
            public int Unknown_36;
            public float Maximum_Budget;
            public float Current_Budget;
            public int Unknown_48;
            public int Unknown_52;
            #endregion
        }
        /// <summary>
        /// Read, writes, and allows editing of individual placement blocks.
        /// </summary>        
        public class PlacementBlock : Chunk
        {
            //Structure of placement block.
            #region Structure
            public short Block_Type; //0
            [Size(10)]
            public byte[] Unused_4; //4
            public int Tags_Index; //12
            public float posX; //16
            public float posY; //20
            public float posZ; //24
            public float posYaw; //28
            public float posPitch; //32
            public float posRoll; //36
            public float Unknown_Rot_1; //40
            public float Unknown_Rot_2; //44
            public float Unknown_Rot_3; //48
            [Size(8)]
            public byte[] Unused_52; //52
            public byte Unknown_60; //60 
            public byte Unknown_61; //61   
            public byte flags; //62
            public byte team; //63
            public byte spareClips; //64
            public byte respawnTime; //65
            public short Unknown66; //66
            [Size(16)]
            public byte[] Unused_68; //68
            #endregion
            //Defines placement block types
            public enum blockType : short
            {
                Player_Spawn = 9,
                Reserved = 41,
                Added = 131,
                Original = 137,
                Edited = 139,
                NULL = 0,
            }
            //Property's for values to allow editing of later.
            #region Properties
            [Browsable(true)]
            [Category("Block Type"), System.ComponentModel.Description("What kind of placement block is this.")]
            public blockType BlockType
            {
                get { return (blockType)Block_Type; }
                set { Block_Type = (short)value; }
            }

            [Category("Spawn Location"), System.ComponentModel.Description("X co-ordinat object will spawn at.")]
            public float X
            {
                get { return posX; }
                set { posX = value; }
            }

            [Category("Spawn Location"), System.ComponentModel.Description("Y co-ordinat object will spawn at.")]
            public float Y
            {
                get { return posY; }
                set { posY = value; }
            }

            [Category("Spawn Location"), System.ComponentModel.Description("Z co-ordinat object will spawn at.")]
            public float Z
            {
                get { return posZ; }
                set { posZ = value; }
            }

            [Category("Object Rotation"), System.ComponentModel.Description("Y axis rotation of object")]
            public float Yaw
            {
                get { return posYaw; }
                set { posYaw = value; }
            }

            [Category("Object Rotation"), System.ComponentModel.Description("P axis rotation of object")]
            public float Pitch
            {
                get { return posPitch; }
                set { posPitch = value; }
            }

            [Category("Object Rotation"), System.ComponentModel.Description("R axis rotation of object")]
            public float Roll
            {
                get { return posRoll; }
                set { posRoll = value; }
            }

            [Category("Respawn Settings")]
            public byte Flags
            {
                get { return flags; }
                set { flags = value; }
            }

            [Category("Respawn Settings"), System.ComponentModel.Description("Team this object belongs to.")]
            public byte Team
            {
                get { return team; }
                set { team = value; }
            }

            [Category("Respawn Settings"), System.ComponentModel.Description("Amount of spare clips does this object has.")]
            public byte SpareClips
            {
                get { return spareClips; }
                set { spareClips = value; }
            }

            [Category("Respawn Settings"), System.ComponentModel.Description("How long this object takes to respawn.")]
            public byte RespawnTime
            {
                get { return respawnTime; }
                set { respawnTime = value; }
            }

            [Category("Tag Reference"), System.ComponentModel.Description("Index of object to spawn in tag index.")]
            public int TagIndex
            {
                get { return Tags_Index; }
                set { Tags_Index = OnTagIndexChanged(value); }
            }

            #endregion
            //Methods for perfoming operations on placement blocks.
            #region Methods
            /// <summary>
            /// Loads placement block with NULL values.
            /// </summary>
            public void LoadNull()
            {
                base.SetFields();

                Block_Type = 0;
                Unused_4 = new byte[] { 0x00, 0x00, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF };
                Tags_Index = -1;
                posX = 0;
                posY = 0;
                posZ = 0;
                posYaw = 0;
                posPitch = 0;
                posRoll = 0;
                Unknown_Rot_1 = 0;
                Unknown_Rot_2 = 0;
                Unknown_Rot_3 = 0;
                Unused_52 = new byte[] { 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF };
                Unknown_60 = 0;
                Unknown_61 = 0;   
                flags = 12;
                team = 9;
                spareClips = 0;
                respawnTime = 0;
                Unknown66 = 0;
                Unused_68 = new byte[] { 0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
                    0x00,0x00,0x00,0x00};
            }

            protected int OnTagIndexChanged(int index)
            {
                if (index >= 0 && index <= 254)
                    return index; //If valid return new index
                else
                    System.Windows.Forms.MessageBox.Show("Index must be between 0 and 254.", "Invalid index entered.", 
                        System.Windows.Forms.MessageBoxButtons.OK, System.Windows.Forms.MessageBoxIcon.Error); //Tell user index entered was invalid.
                return Tags_Index; //Return original index number
            }
            #endregion
        }
        /// <summary>
        /// Read, writes, and allows editing of individual tag entrys.
        /// </summary>
        public class TagIndexEntry : Chunk
        {
            //Structure of tag index entrys
            #region Structure
            public int ident;
            public byte run_Time_Minimun;
            public byte run_Time_Maximun;
            public byte number_On_Map;
            public byte design_Time_Maximun;
            public float total_Cost;
            #endregion
            //Property's for values to allow editing of later.
            #region Properties
            [Category("Object Info"), System.ComponentModel.Description("What object will spawn."),
        Editor(typeof(Sandbox.Editor.UIIdentSwapper), typeof(System.Drawing.Design.UITypeEditor))]
            public string Ident
            {
                get { return GlobalVariables.Plugin.GetTagNameFromID(ident); }
                set { ident = GlobalVariables.Plugin.GetIDFromTagName(value); }                    
            }
            
            [Category("Spawn Rules"), System.ComponentModel.Description("Minimun number of this object that can spawn at one time ingame")]
            public byte RunTimeMinimun
            {
                get { return run_Time_Minimun; }
                set { run_Time_Minimun = value; }
            }
            [Category("Spawn Rules"), System.ComponentModel.Description("Maximun number of this object that can spawn at one time ingame.")]
            public byte RunTimeMaximun
            {
                get { return run_Time_Maximun; }
                set { run_Time_Maximun = value; }
            }
            [Category("Spawn Rules"), System.ComponentModel.Description("Number of this item on the map.")]
            public byte NumberOnMap
            {
                get { return number_On_Map; }
                set { number_On_Map = value; }
            }
            [Category("Spawn Rules"), System.ComponentModel.Description("Maximun number of this object that can be placed at one time while creating map.")]
            public byte DesignTimeMaximun
            {
                get { return design_Time_Maximun; }
                set { design_Time_Maximun = value; }
            }
            [Category("Object Info"), System.ComponentModel.Description("How much per object does it cost to spawn.")]
            public float TotalCost
            {
                get { return total_Cost; }
                set { total_Cost = value; }
            }
            #endregion
            //Methods for perfoming operations on index entrys
            #region Methods
            /// <summary>
            /// Loads placement block with NULL values.
            /// </summary>
            public void LoadNull()
            {
                base.SetFields();

                ident = -1;
                run_Time_Maximun = 0;
                run_Time_Minimun = 0;
                number_On_Map = 0;
                design_Time_Maximun = 255;
                total_Cost = -1;
            }
            #endregion
        }
    }
}
